/*
Layer III routines adopted from the ISO MPEG Audio Subgroup Software Simulation
Group's public c source for its MPEG audio decoder. These routines were in the
file "decoder.c". Rearrangement of the routines as member functions of a
layer III decoder object, and optimizations by Jeff Tsay
(ctsay@pasteur.eecs.berkeley.edu).

If you want to help, figure out how to implement a 9 point IDCT
in n log n time, and e-mail me.

####
I put in the 9 point IDCT by Mikko Tommila from the Windows version 1.81
and cut some time in dequantize_sample.
Fixed one bug which caused maplay to dump a core with certain .mp3 files
Some optimizations regarding loops.
Downmixing now in stereo method.

Hendrik Pagenhardt (pagenhar@sunpool.cs.uni-magdeburg.de)

Last modified : 06/04/97 */

#include <math.h>
#include <string.h>

#include "all.h"
#include "l3type.h"
#include "obuffer.h"
#include "bit_res.h"
#include "header.h"
#include "huffman.h"
#include "layer3.h"

LayerIII_Decoder::LayerIII_Decoder(char *huffdec_path0,
				   Ibitstream *stream0,
                                   Header *header0,
                                   SynthesisFilter *filtera,
                                   SynthesisFilter *filterb,
				   Obuffer *buffer0,
                                   enum e_channels which_ch0)
{
  huffdec_path = huffdec_path0;
  stream = stream0;
  header = header0;
  filter1 = filtera;
  filter2 = filterb;
  buffer = buffer0;
  which_channels = which_ch0;

  frame_start = 0;
  channels = (header->mode() == single_channel) ? 1 : 2;
  max_gr      = 2; // (header->version() == MPEG1) ? 2 : 1;

  if (channels == 2) {

  	  switch (which_channels) {

     case left:
     case downmix:
     first_channel = last_channel = 0;
     break;

     case right:
     first_channel = last_channel = 1;
     break;

     case both:
     first_channel  = 0;
     last_channel   = 1;
     break;

	  default:
     // uh oh
     break;

     }
  } else {

    first_channel = last_channel = 0;
  }

  int32 i,j,k;

  for(i=0;i<2;i++)
    for(j=0;j<SBLIMIT;j++)
      for(k=0;k<SSLIMIT;k++)
	prevblck[i][j][k]=0.0f;

  br = new Bit_Reserve();
  si = new III_side_info_t;

  memset(si,0,sizeof(III_side_info_t));
}

LayerIII_Decoder::~LayerIII_Decoder()
{
   delete br;
   delete si;
}

void LayerIII_Decoder::seek_notify()
{
  frame_start = 0;

  int32 i,j,k;

  for(i=0;i<2;i++)
    for(j=0;j<SBLIMIT;j++)
      for(k=0;k<SSLIMIT;k++)
	prevblck[i][j][k]=0.0f;

  delete br;
  br = new Bit_Reserve;
}


struct {
  int32 l[5];
  int32 s[3];} sfbtable = {{0, 6, 11, 16, 21},
			   {0, 6, 12}};

int32 slen[2][16] = {{0, 0, 0, 0, 3, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4},
                     {0, 1, 2, 3, 0, 1, 2, 3, 1, 2, 3, 1, 2, 3, 2, 3}};

void LayerIII_Decoder::get_scale_factors(int32 ch, int32 gr)
{
  int32 sfb, window;
  gr_info_s *gr_info = &(si->ch[ch].gr[gr]);
  int32 scale_comp = gr_info->scalefac_compress;
  int32 length0 = slen[0][scale_comp];
  int32 length1 = slen[1][scale_comp];

  if (gr_info->window_switching_flag && (gr_info->block_type == 2)) {
    if (gr_info->mixed_block_flag) { /* MIXED */ /* NEW - ag 11/25 */
      for (sfb = 0; sfb < 8; sfb++)
	scalefac[ch].l[sfb] = br->hgetbits(slen[0][gr_info->scalefac_compress]);
      for (sfb = 3; sfb < 6; sfb++)
	for (window=0; window<3; window++)
	  scalefac[ch].s[window][sfb] = br->hgetbits(slen[0][gr_info->scalefac_compress]);
      for (sfb = 6; sfb < 12; sfb++)
	for (window=0; window<3; window++)
	  scalefac[ch].s[window][sfb] = br->hgetbits(slen[1][gr_info->scalefac_compress]);
      for (sfb=12,window=0; window<3; window++)
	scalefac[ch].s[window][sfb] = 0;
    }
    else {  /* SHORT*/

      scalefac[ch].s[0][0] = br->hgetbits(length0);
      scalefac[ch].s[1][0] = br->hgetbits(length0);
      scalefac[ch].s[2][0] = br->hgetbits(length0);
      scalefac[ch].s[0][1] = br->hgetbits(length0);
      scalefac[ch].s[1][1] = br->hgetbits(length0);
      scalefac[ch].s[2][1] = br->hgetbits(length0);
      scalefac[ch].s[0][2] = br->hgetbits(length0);
      scalefac[ch].s[1][2] = br->hgetbits(length0);
      scalefac[ch].s[2][2] = br->hgetbits(length0);
      scalefac[ch].s[0][3] = br->hgetbits(length0);
      scalefac[ch].s[1][3] = br->hgetbits(length0);
      scalefac[ch].s[2][3] = br->hgetbits(length0);
      scalefac[ch].s[0][4] = br->hgetbits(length0);
      scalefac[ch].s[1][4] = br->hgetbits(length0);
      scalefac[ch].s[2][4] = br->hgetbits(length0);
      scalefac[ch].s[0][5] = br->hgetbits(length0);
      scalefac[ch].s[1][5] = br->hgetbits(length0);
      scalefac[ch].s[2][5] = br->hgetbits(length0);
      scalefac[ch].s[0][6] = br->hgetbits(length1);
      scalefac[ch].s[1][6] = br->hgetbits(length1);
      scalefac[ch].s[2][6] = br->hgetbits(length1);
      scalefac[ch].s[0][7] = br->hgetbits(length1);
      scalefac[ch].s[1][7] = br->hgetbits(length1);
      scalefac[ch].s[2][7] = br->hgetbits(length1);
      scalefac[ch].s[0][8] = br->hgetbits(length1);
      scalefac[ch].s[1][8] = br->hgetbits(length1);
      scalefac[ch].s[2][8] = br->hgetbits(length1);
      scalefac[ch].s[0][9] = br->hgetbits(length1);
      scalefac[ch].s[1][9] = br->hgetbits(length1);
      scalefac[ch].s[2][9] = br->hgetbits(length1);
      scalefac[ch].s[0][10] = br->hgetbits(length1);
      scalefac[ch].s[1][10] = br->hgetbits(length1);
      scalefac[ch].s[2][10] = br->hgetbits(length1);
      scalefac[ch].s[0][11] = br->hgetbits(length1);
      scalefac[ch].s[1][11] = br->hgetbits(length1);
      scalefac[ch].s[2][11] = br->hgetbits(length1);
      scalefac[ch].s[0][12] = 0;
      scalefac[ch].s[1][12] = 0;
      scalefac[ch].s[2][12] = 0;
    }
  }
  else {   /* LONG types 0,1,3 */

    if ((si->ch[ch].scfsi[0] == 0) || (gr == 0)) {
      scalefac[ch].l[0] = br->hgetbits(length0);
      scalefac[ch].l[1] = br->hgetbits(length0);
      scalefac[ch].l[2] = br->hgetbits(length0);
      scalefac[ch].l[3] = br->hgetbits(length0);
      scalefac[ch].l[4] = br->hgetbits(length0);
      scalefac[ch].l[5] = br->hgetbits(length0);
    }
    if ((si->ch[ch].scfsi[1] == 0) || (gr == 0)) {
      scalefac[ch].l[6] = br->hgetbits(length0);
      scalefac[ch].l[7] = br->hgetbits(length0);
      scalefac[ch].l[8] = br->hgetbits(length0);
      scalefac[ch].l[9] = br->hgetbits(length0);
      scalefac[ch].l[10] = br->hgetbits(length0);
    }
    if ((si->ch[ch].scfsi[2] == 0) || (gr == 0)) {
      scalefac[ch].l[11] = br->hgetbits(length1);
      scalefac[ch].l[12] = br->hgetbits(length1);
      scalefac[ch].l[13] = br->hgetbits(length1);
      scalefac[ch].l[14] = br->hgetbits(length1);
      scalefac[ch].l[15] = br->hgetbits(length1);
    }
    if ((si->ch[ch].scfsi[3] == 0) || (gr == 0)) {
      scalefac[ch].l[16] = br->hgetbits(length1);
      scalefac[ch].l[17] = br->hgetbits(length1);
      scalefac[ch].l[18] = br->hgetbits(length1);
      scalefac[ch].l[19] = br->hgetbits(length1);
      scalefac[ch].l[20] = br->hgetbits(length1);
    }

    scalefac[ch].l[21] = 0;
    scalefac[ch].l[22] = 0;
  }
}

// ####
struct _sfBandIndex {
  int32 l[23];
  int32 s[14];} sfBandIndex[3] =
		{{{0,4,8,12,16,20,24,30,36,44,52,62,74,90,110,134,162,196,238,288,342,418,576},
		  {0,4,8,12,16,22,30,40,52,66,84,106,136,192}},
		 {{0,4,8,12,16,20,24,30,36,42,50,60,72,88,106,128,156,190,230,276,330,384,576},
		  {0,4,8,12,16,22,28,38,50,64,80,100,126,192}},
		 {{0,4,8,12,16,20,24,30,36,44,54,66,82,102,126,156,194,240,296,364,448,550,576},
		  {0,4,8,12,16,22,30,42,58,78,104,138,180,192}}};

void LayerIII_Decoder::hufman_decode(int32 ch, int32 gr)
{
  int32 i, x, y;
  int32 v, w;
  register gr_info_s *sip = &(si->ch[ch].gr[gr]); // #### helps the compiler
  int32 part2_3_end = part2_start + sip->part2_3_length;
  int32 num_bits;
  int32 region1Start;
  int32 region2Start;
  int32 sf_index;
// #### int32 ssindex, sbindex; // #### not neccessary any longer
  struct huffcodetab *h;

  static BOOL huffman_init = FALSE;

  if (!huffman_init) {
    if (read_decoder_table(huffdec_path))
      huffman_init = TRUE;
    else
      exit(1L);
  }

  /* Find region boundary for short block case. */

  if ( (sip->window_switching_flag) &&
       (sip->block_type == 2) ) {

    /* Region2. */

    region1Start = 36;  /* sfb[9/3]*3=36 */
    region2Start = 576; /* No Region2 for short block case. */
  }

  else {          /* Find region boundary for long block case. */

    sf_index = header->sample_frequency();

    region1Start = sfBandIndex[sf_index].l[sip->region0_count + 1]; /* MI */
    region2Start = sfBandIndex[sf_index].l[sip->region0_count +
					  sip->region1_count + 2]; /* MI */
  }

// #### sbindex = 0; ssindex = 0;
  register int32 *isp = &is[0][0];
  /* Read bigvalues area. */
  for (i=0; i<(sip->big_values<<1); i+=2) {
    if      (i<region1Start) h = &ht[sip->table_select[0]];
    else if (i<region2Start) h = &ht[sip->table_select[1]];
    else                h = &ht[sip->table_select[2]];

    huffman_decoder(h, &x, &y, &v, &w, br);

/* ####
    is[sbindex][ssindex]   = x;
    is[sbindex][ssindex+1] = y;

    ssindex += 2;
    if (ssindex >= SSLIMIT) {
      ssindex = 0;
      sbindex++;
    }
*/
    *isp = x;
    *(isp + 1) = y;
    isp += 2;
  }

  /* Read count1 area. */
  h = &ht[sip->count1table_select+32];
  num_bits = br->hsstell();

  register int32 *isp_end = &is[0][0] + SBLIMIT*SSLIMIT;
  while ((num_bits < part2_3_end) && ( isp < isp_end/*sbindex < SBLIMIT*/)) {

    huffman_decoder(h, &x, &y, &v, &w, br);

/* ####
    is[sbindex][ssindex] = v;
    is[sbindex][ssindex + 1] = w;

    ssindex += 2;
    if (ssindex >= SSLIMIT) {
      ssindex = 0;
      sbindex++;
    }
*/
    *isp = v;
    *(isp + 1) = w;
    isp += 2;

/* ####
    if (sbindex < SBLIMIT) {
      is[sbindex][ssindex]   = x;
      is[sbindex][ssindex+1] = y;
    }
    ssindex += 2;
    if (ssindex >= SSLIMIT) {
      ssindex = 0;
      sbindex++;
    }
*/
    if (isp < isp_end) {
    	*isp = x;
    	*(isp + 1) = y;
    }
    isp += 2;

    num_bits = br->hsstell();
   }

// #### change taken over from win32 version 1.81 and modified
   if (num_bits > part2_3_end) {
	br->rewindNbits(num_bits - part2_3_end);
/* ####
	ssindex -= 4;
	if (ssindex < 0) {
		ssindex += SSLIMIT;
		sbindex--;
	}
*/
	isp -= 4;
   }

  num_bits = br->hsstell();

  /* Dismiss stuffing Bits */
  if (num_bits < part2_3_end)
    br->hgetbits(part2_3_end - num_bits);

  /* Zero out rest. */
  /*for (; i<SSLIMIT*SBLIMIT; i++)
    is[i/SSLIMIT][i%SSLIMIT] = 0; */

// ####
   if (isp < isp_end)
   {
	register int32 *ssi = isp;
	register int32 *ssi_end = isp + (isp_end - isp)%SSLIMIT;
	for (; ssi < ssi_end; *ssi++ = 0) ;
	ssi_end = isp_end;
	for(; ssi < ssi_end; ssi += SSLIMIT) {
		ssi[ 0] = 0; ssi[ 1] = 0; ssi[ 2] = 0; ssi[ 3] = 0; 
		ssi[ 4] = 0; ssi[ 5] = 0; ssi[ 6] = 0; ssi[ 7] = 0; 
		ssi[ 8] = 0; ssi[ 9] = 0; ssi[10] = 0; ssi[11] = 0; 
		ssi[12] = 0; ssi[13] = 0; ssi[14] = 0; ssi[15] = 0; 
		ssi[16] = 0; ssi[17] = 0;
	}
   }
}

int pretab[22] = {0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,2,2,3,3,3,2,0};

void LayerIII_Decoder::dequantize_sample(real xr[SBLIMIT][SSLIMIT],
						int32 ch, int32 gr)
{
	gr_info_s *gr_info = &(si->ch[ch].gr[gr]);

	int32 cb=0;
// ####
	struct _sfBandIndex *sfbifp = &sfBandIndex[header->sample_frequency()];
	int32 next_cb_boundary, cb_begin, cb_width;
	real temp;
	int32 t_index;
   int32 i;
   static BOOL Dequant_init=FALSE;
   static real TO_FOUR_THIRDS[10240]; // #### 1024 was not enough!
   static real two_to_negative_half_pow[40];
// ####
	static double sqrt05;
	static double sqrtsqrt2;
	static real sqrt05pow[128];
	static real global_scale[256];

   if (!Dequant_init) {

	for(i = 0; i<10240; i++)
	  		TO_FOUR_THIRDS[i] = (real) pow( (double) i, 4.0/3.0);

	for(i=0; i<40; i++)
		two_to_negative_half_pow[i] = (real) pow(2.0, -0.5 * (double) i);

	sqrt05 = sqrt(2.0)/2.0;
	sqrtsqrt2 = sqrt(sqrt(2.0));
	for(i = 0; i<128; i++)
		sqrt05pow[ i] = pow( sqrt05, i);

	for(i = 0; i<256; i++) 
		global_scale[i] = (real) pow( sqrtsqrt2, i - 210.0);

	Dequant_init = TRUE;
   }

	/* choose correct scalefactor band per block type, initalize boundary */

	if (gr_info->window_switching_flag && (gr_info->block_type == 2) )
		if (gr_info->mixed_block_flag)
			next_cb_boundary=sfbifp->l[1];  /* LONG blocks: 0,1,3 */
		else {
		   cb_width = sfbifp->s[1];
		   next_cb_boundary = cb_width * 3; // #### (cb_width << 2) - cb_width;
	 	   cb_begin = 0;
		}
	else
		next_cb_boundary=sfbifp->l[1];  /* LONG blocks: 0,1,3 */

   /* Compute overall (global) scaling. */

// ####	temp = (real) pow( 2.0 , (0.25 * (gr_info->global_gain - 210.0)));
//	temp = (real) pow( sqrtsqrt2, gr_info->global_gain - 210.0);
	temp = global_scale[gr_info->global_gain];

// ####
	int32 *isp = is[0];
	real *xrp = xr[0];
	/* apply formula per block type */

		uint32 index = 0;
		do {

			if (index == next_cb_boundary)  { /* Adjust critical band boundary */
				if (gr_info->window_switching_flag && (gr_info->block_type == 2)) {
					if (gr_info->mixed_block_flag)  {

						if (index == sfbifp->l[8])  {
							cb = 3;
							register int32 *t1;
							cb_begin = 3 * *(t1 = &sfbifp->s[3]);
							next_cb_boundary = 3 * t1[1];
							cb_width = t1[1] - t1[0];
						}

						else 	if (index < sfbifp->l[8])
								next_cb_boundary = sfbifp->l[(++cb)+1];
							else {
								register int32 *t1;
								cb_begin = 3 * *(t1 = &sfbifp->s[++cb]);
								next_cb_boundary = 3 * t1[1];
								cb_width = t1[1] - t1[0];
							}
					}
					else  {
						register int32 *t1;
						cb_begin = 3 * *(t1 = &sfbifp->s[++cb]);
						next_cb_boundary = 3 * t1[1];
						cb_width = t1[1] - t1[0];
					}
				}
				else /* long blocks */
					next_cb_boundary = sfbifp->l[(++cb)+1];
			}

			/* Do long/short dependent scaling operations. */

			register real t1, t2;
			register int32 t3;
			if ((t3 = *isp++) < 0 ? t3 = -t3, True: False)
				t1   = - temp * TO_FOUR_THIRDS[ t3];
			else
				t1   =   temp * TO_FOUR_THIRDS[ t3];
			if (gr_info->window_switching_flag &&
				 (((gr_info->block_type == 2) && (gr_info->mixed_block_flag == 0)) ||
				  ((gr_info->block_type == 2) && gr_info->mixed_block_flag && (index >= 2 * SSLIMIT)) ))
			{

				t_index = (index - cb_begin) / cb_width;
// #### gets rid of this pow() call
				t2 = t1 * sqrt05pow[ (gr_info->subblock_gain[t_index] << 2)
						+ ((1 + gr_info->scalefac_scale)
						* scalefac[ch].s[t_index][cb])];
//				xr[sb][ss] *= pow(2.0, ((-2.0 * gr_info->subblock_gain[t_index])
//						-(0.5 * (1.0 + gr_info->scalefac_scale)
//						* scalefac[ch].s[t_index][cb])));
			}
			else {   /* LONG block types 0,1,3 & 1st 2 subbands of switched blocks */
/*				xr[sb][ss] *= pow(2.0, -0.5 * (1.0+gr_info->scalefac_scale)
							 * (scalefac[ch].l[cb]
							 + gr_info->preflag * pretab[cb])); */
/*				xr[sb][ss] *= two_pow2(gr_info->scalefac_scale, gr_info->preflag,
                                   pretab[cb], scalefac[ch].l[cb]);*/
// #### change taken from win32 version 1.81
				uint32 idx = scalefac[ch].l[cb];

				if (gr_info->preflag)
					idx += pretab[cb];
				idx = idx << gr_info->scalefac_scale;
				t2 = t1 * two_to_negative_half_pow[idx];
			}
			*xrp++ = t2;
		} while (++index < SSLIMIT*SBLIMIT);
}

void LayerIII_Decoder::reorder (real xr[SBLIMIT][SSLIMIT], int32 ch, int32 gr)
{

	gr_info_s *gr_info = &(si->ch[ch].gr[gr]);
// ####
	struct _sfBandIndex *sfbifp = &sfBandIndex[header->sample_frequency()];
	int32 sfb, sfb_start, sfb_lines;
	int32 freq, src_line, des_line;

	if (gr_info->window_switching_flag && (gr_info->block_type == 2)) {

		for(register uint32 sb=0;sb<SBLIMIT;sb++)
			re_hybridOut[sb][0]  = re_hybridOut[sb][1]  = re_hybridOut[sb][2]  =
			re_hybridOut[sb][3]  = re_hybridOut[sb][4]  = re_hybridOut[sb][5]  =
			re_hybridOut[sb][6]  = re_hybridOut[sb][7]  = re_hybridOut[sb][8]  =
			re_hybridOut[sb][9]  = re_hybridOut[sb][10] = re_hybridOut[sb][11] =
			re_hybridOut[sb][12] = re_hybridOut[sb][13] = re_hybridOut[sb][14] =
			re_hybridOut[sb][15] = re_hybridOut[sb][16] = re_hybridOut[sb][17] =
			0.0f;

		if (gr_info->mixed_block_flag) {
			/* NO REORDER FOR LOW 2 SUBBANDS */
// ####
	    memcpy( re_hybridOut[0], xr[0], SSLIMIT * sizeof( real));
/*
			   for (register uint32 ss=0 ; ss < SSLIMIT ; ss+=3) {
				re_hybridOut[0][ss]   = xr[0][ss];
				re_hybridOut[0][ss+1] = xr[0][ss+1];
				re_hybridOut[0][ss+2] = xr[0][ss+2];
            }
*/
// ####
	    memcpy( re_hybridOut[1], xr[1], SSLIMIT * sizeof( real));
/*            for (register uint32 ss=0; ss < SSLIMIT ; ss+=3) {
		re_hybridOut[1][ss]   = xr[1][ss];
		re_hybridOut[1][ss+1] = xr[1][ss+1];
		re_hybridOut[1][ss+2] = xr[1][ss+2];
            }
*/
			/* REORDERING FOR REST SWITCHED SHORT */
				for(sfb=3,sfb_start=sfbifp->s[3],
	  				 sfb_lines=sfbifp->s[4] - sfb_start;
	  				 sfb < 13; sfb++,sfb_start=sfbifp->s[sfb],
				    (sfb_lines=sfbifp->s[sfb+1] - sfb_start))
            {
					   int32 sfb_start3 = sfb_start * 3; // #### (sfb_start << 2) - sfb_start;

						for(freq=0;freq<sfb_lines;freq++) {
							int32 freq3 = freq * 3; // #### (freq << 2) - freq;
/* ####
							src_line = sfb_start3 + freq;
							des_line = sfb_start3 + freq3;
							re_hybridOut[des_line/SSLIMIT][des_line%SSLIMIT] =
												xr[src_line/SSLIMIT][src_line%SSLIMIT];
							src_line += sfb_lines;
							des_line++;
							re_hybridOut[des_line/SSLIMIT][des_line%SSLIMIT] =
												xr[src_line/SSLIMIT][src_line%SSLIMIT];
							src_line += sfb_lines;
							des_line++;
							re_hybridOut[des_line/SSLIMIT][des_line%SSLIMIT] =
												xr[src_line/SSLIMIT][src_line%SSLIMIT];
*/
// ####
							register real *xrp = &xr[0][0] + sfb_start3 + freq;
							register real *rep = &re_hybridOut[0][0] + sfb_start3 + freq3;
							*rep++ = *xrp; xrp += sfb_lines;
							*rep++ = *xrp; xrp += sfb_lines;
							*rep = *xrp;
					   }
            }
		}
		else {  /* pure short */
			for(sfb=0,sfb_start=0,sfb_lines=sfbifp->s[1];
				sfb < 13; sfb++,sfb_start=sfbifp->s[sfb],
				(sfb_lines=sfbifp->s[sfb+1] - sfb_start))
            {
					   int32 sfb_start3 = sfb_start * 3; // #### (sfb_start << 2) - sfb_start;

						for(freq=0;freq<sfb_lines;freq++) {
							int32 freq3 = freq * 3; // #### (freq << 2) - freq;
/* ####
							src_line = sfb_start3 + freq;
							des_line = sfb_start3 + freq3;
							re_hybridOut[des_line/SSLIMIT][des_line%SSLIMIT] =
												xr[src_line/SSLIMIT][src_line%SSLIMIT];
							src_line += sfb_lines;
							des_line++;
							re_hybridOut[des_line/SSLIMIT][des_line%SSLIMIT] =
												xr[src_line/SSLIMIT][src_line%SSLIMIT];
							src_line += sfb_lines;
							des_line++;
							re_hybridOut[des_line/SSLIMIT][des_line%SSLIMIT] =
												xr[src_line/SSLIMIT][src_line%SSLIMIT];
*/
// ####
							register real *xrp = &xr[0][0] + sfb_start3 + freq;
							register real *rep = &re_hybridOut[0][0] + sfb_start3 + freq3;
							*rep++ = *xrp; xrp += sfb_lines;
							*rep++ = *xrp; xrp += sfb_lines;
							*rep = *xrp;
					   }
            }
		}
	}
	else {   /*long blocks */
		memcpy( re_hybridOut, xr, SBLIMIT * SSLIMIT * sizeof( real));
/*
		for (sb=0 ; sb < SBLIMIT ; sb++)
         		for (ss = 0; ss < SSLIMIT; ss+=3) {
				re_hybridOut[sb][ss]   = xr[sb][ss];
				re_hybridOut[sb][ss+1] = xr[sb][ss+1];
				re_hybridOut[sb][ss+2] = xr[sb][ss+2];
			}
*/
	}
}

//#define PI   3.141593
#define PI12      0.261799387f
#define PI36      0.087266462f
#define COSPI3    0.500000000f
#define COSPI6    0.866025403f
#define DCTODD1   0.984807753f
#define DCTODD2  -0.342020143f
#define DCTODD3  -0.642787609f
#define DCTEVEN1  0.939692620f
#define DCTEVEN2 -0.173648177f
#define DCTEVEN3 -0.766044443f

real TAN12[16]={0.0, 0.26794919, 0.57735027, 1.0,
                1.73205081, 3.73205081, 9.9999999e10 /*unbounded*/, -3.73205081,
                -1.73205081, -1.0, -0.57735027, -0.26794919,
                0.0, 0.26794919, 0.57735027, 1.0};

void LayerIII_Decoder::stereo(int32 gr)
{
	gr_info_s *gr_info = &(si->ch[0].gr[gr]);
// ####
	struct _sfBandIndex *sfbifp = &sfBandIndex[header->sample_frequency()];
	int32 mode_ext = header->mode_extension();

	int32 ms_stereo = (header->mode() == joint_stereo) && (mode_ext & 0x2);
	int32 i_stereo  = (header->mode() == joint_stereo) && (mode_ext & 0x1);
	int32 sfb;
	register int32 i,j,sb,ss;
	int32 is_pos[576];
	int32 lines, temp, temp2;
	real is_ratio[576];

	if  (channels == 1) { /* mono , bypass xr[0][][] to lr[0][][]*/
		memcpy(lr[0], ro[0], sizeof(lr[0][0][0]) * SBLIMIT * SSLIMIT);
/* ####		for(sb=0;sb<SBLIMIT;sb++)
			for(ss=0;ss<SSLIMIT;ss+=3) {
				lr[0][sb][ss]   = ro[0][sb][ss];
				lr[0][sb][ss+1] = ro[0][sb][ss+1];
				lr[0][sb][ss+2] = ro[0][sb][ss+2];
         }
*/
   } else {

 	/* initialization */
// ####
   for (register int32 *isp = is_pos + 576; isp > is_pos; isp -= 8) {
      isp[-1] = isp[-2] = isp[-3] = isp[-4] =
      isp[-5] = isp[-6] = isp[-7] = isp[-8] = 7;
   }
	if (i_stereo)
	{  if (gr_info->window_switching_flag && (gr_info->block_type == 2))
		{  if( gr_info->mixed_block_flag )
			{  int32 max_sfb = 0;

				for (j=0; j<3; j++)
				{  int32 sfbcnt;
					sfbcnt = 2;
					for( sfb=12; sfb >=3; sfb-- )
					{
						i = sfbifp->s[sfb];
						lines = sfbifp->s[sfb+1] - i;
						real *ro1p = (real *)ro[1] + i * 3 + (j+1) * lines - 1;
						int32 l = lines;
						while ( l-- > 0 ) {
// ####						   if (ro[1][i/SSLIMIT][i%SSLIMIT] != 0.0f)
						   if (*(ro1p--) != 0.0f)
							{	sfbcnt = sfb;
								sfb = -10;
								l = -10;
							}
						}
					}
					sfb = sfbcnt + 1;

					if (sfb > max_sfb)
						max_sfb = sfb;

					while(sfb < 12)
					{	temp = sfbifp->s[sfb];
						sb   = sfbifp->s[sfb+1] - temp;
						i    = temp * 3 /* #### (temp << 2) - temp */ + j * sb;

						for ( ; sb > 0; sb--)
						{  is_pos[i] = scalefac[1].s[j][sfb];
							if (is_pos[i] != 7)
								is_ratio[i] = TAN12[is_pos[i]];

							i++;
						}
						sfb++;
					}
					sfb = sfbifp->s[10];
					sb  = sfbifp->s[11] - sfb;
					sfb = sfb * 3 /* #### (sfb << 2) - sfb */ + j * sb;
					temp  = sfbifp->s[11];
					sb = sfbifp->s[12] - temp;
					i = temp * 3 /* #### (temp << 2) - temp */ + j * sb;

					for ( ; sb > 0; sb-- )
					{  is_pos[i] = is_pos[sfb];
						is_ratio[i] = is_ratio[sfb];
						i++;
					}
				 }
				 if (max_sfb <= 3)
				 {  i = 2;
					 ss = 17;
					 sb = -1;
					 while (i >= 0)
					 {  if (ro[1][i][ss] != 0.0f)
						 {
							sb = (i<<4) + (i<<1) + ss;
							 i = -1;
						 } else
						 {  ss--;
							 if (ss < 0)
							 {  i--;
								 ss = 17;
							 }
						 }
					 }
					 i = 0;
					 while ( sfbifp->l[i] <= sb )
						 i++;
					 sfb = i;
					 i = sfbifp->l[i];
					 for (; sfb<8; sfb++)
					 {  sb = sfbifp->l[sfb+1]-sfbifp->l[sfb];
						 for (; sb>0; sb--)
						 {  is_pos[i] = scalefac[1].l[sfb];
							 if (is_pos[i] != 7)
								is_ratio[i] = TAN12[is_pos[i]];

							 i++;
						 }
					 }
				}
			} else
			{  for (j=0; j<3; j++)
				{  int32 sfbcnt;
					sfbcnt = -1;
					for( sfb=12; sfb >=0; sfb-- )
					{
						temp = sfbifp->s[sfb];
						lines = sfbifp->s[sfb+1] - temp;
						real *ro1p = (real *)ro[1] + temp * 3 + (j+1) * lines - 1;
						int32 l = lines;
						while ( l-- > 0 ) {
// ####						   if (ro[1][i/SSLIMIT][i%SSLIMIT] != 0.0f)
						   if (*(ro1p--) != 0.0f)
							{  sfbcnt = sfb;
								sfb = -10;
								l = -10;
							}
						}
					}
					sfb = sfbcnt + 1;
					while( sfb<12 )
					{
						temp = sfbifp->s[sfb];
						sb   = sfbifp->s[sfb+1] - temp;
						i    = temp * 3 /* #### (temp << 2) - temp */ + j * sb;
						for ( ; sb > 0; sb--)
						{  is_pos[i] = scalefac[1].s[j][sfb];
							if (is_pos[i] != 7)
								is_ratio[i] = TAN12[is_pos[i]];

							i++;
						}
						sfb++;
					}

					temp = sfbifp->s[10];
					temp2= sfbifp->s[11];
					sb   = temp2 - temp;
					sfb  = temp * 3 /* #### (temp << 2) - temp */ + j * sb;
					sb   = sfbifp->s[12] - temp2;
					i    = temp2 * 3 /* #### (temp2 << 2) - temp2 */ + j * sb;

					for (; sb>0; sb--)
					{  is_pos[i] = is_pos[sfb];
						is_ratio[i] = is_ratio[sfb];
						i++;
					}
				}
			}
		} else // ms-stereo
		{  i = 31;
			ss = 17;
			sb = 0;
			while (i >= 0)
			{  if (ro[1][i][ss] != 0.0f)
				{
					sb = (i<<4) + (i<<1) + ss;
					i = -1;
				} else
				{  ss--;
					if ( ss < 0 )
					{  i--;
						ss = 17;
					}
				}
			}
			i = 0;
			while ( sfbifp->l[i] <= sb )
				i++;
			sfb = i;
			i = sfbifp->l[i];
			for ( ; sfb<21; sfb++ )
			{  sb = sfbifp->l[sfb+1] - sfbifp->l[sfb];
				for ( ; sb > 0; sb--)
				{  is_pos[i] = scalefac[1].l[sfb];
					if ( is_pos[i] != 7 )
						is_ratio[i] = TAN12[is_pos[i]];

					i++;
				}
			}
			sfb = sfbifp->l[20];
			for ( sb = 576 - sfbifp->l[21]; sb > 0; sb-- )
			{  is_pos[i] = is_pos[sfb];
				is_ratio[i] = is_ratio[sfb];
				i++;
			}
		}
	}

   	i = 0;
// ####
	if (which_channels == downmix) {
		for(register uint32 sbi=0;sbi<SBLIMIT;sbi++)
			for(register uint32 ssi=0;ssi<SSLIMIT;ssi++) {
				if ( is_pos[i] == 7 ) {
					if (ms_stereo) {
						lr[0][sbi][ssi] = ro[0][sbi][ssi] * 0.707106781f;
					}
					else {
						lr[0][sbi][ssi] = (ro[0][sbi][ssi] + ro[1][sbi][ssi]) * 0.5f;
					}
				}
				else if (i_stereo ) {
					lr[0][sbi][ssi] = ro[0][sbi][ssi] * 0.5f;
				}
/*				else {
					printf("Error in stereo processing\n");
				} */
				i++;
			}
	} else {
		for(register uint32 sbi=0;sbi<SBLIMIT;sbi++)
			for(register uint32 ssi=0;ssi<SSLIMIT;ssi++) {
				if ( is_pos[i] == 7 ) {
					if (ms_stereo) {
						lr[0][sbi][ssi] = (ro[0][sbi][ssi]+ro[1][sbi][ssi])* 0.707106781f;
						lr[1][sbi][ssi] = (ro[0][sbi][ssi]-ro[1][sbi][ssi])* 0.707106781f;
					}
					else {
						lr[0][sbi][ssi] = ro[0][sbi][ssi];
						lr[1][sbi][ssi] = ro[1][sbi][ssi];
					}
				}
				else if (i_stereo ) {
					lr[1][sbi][ssi] = ro[0][sbi][ssi] / (real) (1 + is_ratio[i]);
					lr[0][sbi][ssi] = lr[1][sbi][ssi] * is_ratio[i];
				}
/*				else {
					printf("Error in stereo processing\n");
				} */
				i++;
			}
	}
    } // channels == 2
}

real Ci[8]={-0.6f,-0.535f,-0.33f,-0.185f,-0.095f,-0.041f,-0.0142f,-0.0037f};

void LayerIII_Decoder::antialias(int32 ch, int32 gr)
{
   gr_info_s *gr_info = &(si->ch[ch].gr[gr]);

	real          bu,bd;  /* upper and lower butterfly inputs */
	int32         ss,sb;
 	static real   ca[8],cs[8];
 	static BOOL   antialias_init = FALSE;

	if (!antialias_init) {
		int32 i;
		real    sq;
		for (i=0;i<8;i++) {
			sq=sqrt(1.0f+Ci[i]*Ci[i]);
			cs[i] = 1.0f/sq;
			ca[i] = Ci[i] * cs[i];
		}
		antialias_init = TRUE;
	}

	/* 31 alias-reduction operations between each pair of sub-bands */
	/* with 8 butterflies between each pair                         */

	if  (gr_info->window_switching_flag && (gr_info->block_type == 2) &&
		 !gr_info->mixed_block_flag ) {

// ####
	memcpy( hybridIn, re_hybridOut, sizeof(real) * SBLIMIT * SSLIMIT);
/*   	for(sb=0;sb<SBLIMIT;sb++) {
        	hybridIn[sb][0] = re_hybridOut[sb][0];
        	hybridIn[sb][1] = re_hybridOut[sb][1];
        	hybridIn[sb][2] = re_hybridOut[sb][2];
        	hybridIn[sb][3] = re_hybridOut[sb][3];
        	hybridIn[sb][4] = re_hybridOut[sb][4];
        	hybridIn[sb][5] = re_hybridOut[sb][5];
        	hybridIn[sb][6] = re_hybridOut[sb][6];
        	hybridIn[sb][7] = re_hybridOut[sb][7];
        	hybridIn[sb][8] = re_hybridOut[sb][8];
        	hybridIn[sb][9] = re_hybridOut[sb][9];
        	hybridIn[sb][10]= re_hybridOut[sb][10];
        	hybridIn[sb][11]= re_hybridOut[sb][11];
        	hybridIn[sb][12]= re_hybridOut[sb][12];
        	hybridIn[sb][13]= re_hybridOut[sb][13];
        	hybridIn[sb][14]= re_hybridOut[sb][14];
        	hybridIn[sb][15]= re_hybridOut[sb][15];
       		hybridIn[sb][16]= re_hybridOut[sb][16];
        	hybridIn[sb][17]= re_hybridOut[sb][17];
      }
*/
	} else if (gr_info->window_switching_flag && gr_info->mixed_block_flag &&
	  (gr_info->block_type == 2)) {

// ####
		memcpy( hybridIn, re_hybridOut, sizeof(real) * 10);
/*		hybridIn[0][0] = re_hybridOut[0][0];
		hybridIn[0][1] = re_hybridOut[0][1];
		hybridIn[0][2] = re_hybridOut[0][2];
		hybridIn[0][3] = re_hybridOut[0][3];
      		hybridIn[0][4] = re_hybridOut[0][4];
		hybridIn[0][5] = re_hybridOut[0][5];
		hybridIn[0][6] = re_hybridOut[0][6];
		hybridIn[0][7] = re_hybridOut[0][7];
		hybridIn[0][8] = re_hybridOut[0][8];
		hybridIn[0][9] = re_hybridOut[0][9];
*/
// #### 
{
      register real *o1=&re_hybridOut[0][17], *o2=re_hybridOut[1], *i1=&hybridIn[0][17], *i2=hybridIn[1];
	i1[ 0] = (o1[ 0] * cs[ 0]) - (o2[ 0] * ca[ 0]);
	i2[ 0] = (o2[ 0] * cs[ 0]) + (o1[ 0] * ca[ 0]);
	i1[-1] = (o1[-1] * cs[ 1]) - (o2[ 1] * ca[ 1]);
	i2[ 1] = (o2[ 1] * cs[ 1]) + (o1[-1] * ca[ 1]);
	i1[-2] = (o1[-2] * cs[ 2]) - (o2[ 2] * ca[ 2]);
	i2[ 2] = (o2[ 2] * cs[ 2]) + (o1[-2] * ca[ 2]);
	i1[-3] = (o1[-3] * cs[ 3]) - (o2[ 3] * ca[ 3]);
	i2[ 3] = (o2[ 3] * cs[ 3]) + (o1[-3] * ca[ 3]);
	i1[-4] = (o1[-4] * cs[ 4]) - (o2[ 4] * ca[ 4]);
	i2[ 4] = (o2[ 4] * cs[ 4]) + (o1[-4] * ca[ 4]);
	i1[-5] = (o1[-5] * cs[ 5]) - (o2[ 5] * ca[ 5]);
	i2[ 5] = (o2[ 5] * cs[ 5]) + (o1[-5] * ca[ 5]);
	i1[-6] = (o1[-6] * cs[ 6]) - (o2[ 6] * ca[ 6]);
	i2[ 6] = (o2[ 6] * cs[ 6]) + (o1[-6] * ca[ 6]);
	i1[-7] = (o1[-7] * cs[ 7]) - (o2[ 7] * ca[ 7]);
	i2[ 7] = (o2[ 7] * cs[ 7]) + (o1[-7] * ca[ 7]);
}

// ####
	memcpy( &hybridIn[1][8], &re_hybridOut[1][8], sizeof(real) * 10);
/*      hybridIn[1][8]  = re_hybridOut[1][8];
      hybridIn[1][9]  = re_hybridOut[1][9];
      hybridIn[1][10] = re_hybridOut[1][10];
      hybridIn[1][11] = re_hybridOut[1][11];
      hybridIn[1][12] = re_hybridOut[1][12];
      hybridIn[1][13] = re_hybridOut[1][13];
      hybridIn[1][14] = re_hybridOut[1][14];
      hybridIn[1][15] = re_hybridOut[1][15];
      hybridIn[1][16] = re_hybridOut[1][16];
      hybridIn[1][17] = re_hybridOut[1][17];
*/
// ####
		memcpy( hybridIn[2], re_hybridOut[2], sizeof(real) * SSLIMIT * (SBLIMIT - 2));
/*		for(sb=2;sb<SBLIMIT;sb++) {

        	hybridIn[sb][0] = re_hybridOut[sb][0];
        	hybridIn[sb][1] = re_hybridOut[sb][1];
        	hybridIn[sb][2] = re_hybridOut[sb][2];
        	hybridIn[sb][3] = re_hybridOut[sb][3];
        	hybridIn[sb][4] = re_hybridOut[sb][4];
        	hybridIn[sb][5] = re_hybridOut[sb][5];
        	hybridIn[sb][6] = re_hybridOut[sb][6];
        	hybridIn[sb][7] = re_hybridOut[sb][7];
        	hybridIn[sb][8] = re_hybridOut[sb][8];
        	hybridIn[sb][9] = re_hybridOut[sb][9];
        	hybridIn[sb][10]= re_hybridOut[sb][10];
        	hybridIn[sb][11]= re_hybridOut[sb][11];
        	hybridIn[sb][12]= re_hybridOut[sb][12];
        	hybridIn[sb][13]= re_hybridOut[sb][13];
        	hybridIn[sb][14]= re_hybridOut[sb][14];
        	hybridIn[sb][15]= re_hybridOut[sb][15];
        	hybridIn[sb][16]= re_hybridOut[sb][16];
        	hybridIn[sb][17]= re_hybridOut[sb][17];
      }
*/
	} else {

// ####
	memcpy( hybridIn, re_hybridOut, sizeof(real) * 8);
/*      hybridIn[0][0] = re_hybridOut[0][0];
      hybridIn[0][1] = re_hybridOut[0][1];
      hybridIn[0][2] = re_hybridOut[0][2];
      hybridIn[0][3] = re_hybridOut[0][3];
      hybridIn[0][4] = re_hybridOut[0][4];
      hybridIn[0][5] = re_hybridOut[0][5];
      hybridIn[0][6] = re_hybridOut[0][6];
      hybridIn[0][7] = re_hybridOut[0][7];
*/
// ####
		for(sb=0;sb<31;sb++) {
			register real *o1=&re_hybridOut[sb][17], *o2=re_hybridOut[sb+1],
				 *i1=&hybridIn[sb][17], *i2=hybridIn[sb+1];
			i1[ 0] = (o1[ 0] * cs[ 0]) - (o2[ 0] * ca[ 0]);
			i2[ 0] = (o2[ 0] * cs[ 0]) + (o1[ 0] * ca[ 0]);
			i1[-1] = (o1[-1] * cs[ 1]) - (o2[ 1] * ca[ 1]);
			i2[ 1] = (o2[ 1] * cs[ 1]) + (o1[-1] * ca[ 1]);
			i1[-2] = (o1[-2] * cs[ 2]) - (o2[ 2] * ca[ 2]);
			i2[ 2] = (o2[ 2] * cs[ 2]) + (o1[-2] * ca[ 2]);
			i1[-3] = (o1[-3] * cs[ 3]) - (o2[ 3] * ca[ 3]);
			i2[ 3] = (o2[ 3] * cs[ 3]) + (o1[-3] * ca[ 3]);
			i1[-4] = (o1[-4] * cs[ 4]) - (o2[ 4] * ca[ 4]);
			i2[ 4] = (o2[ 4] * cs[ 4]) + (o1[-4] * ca[ 4]);
			i1[-5] = (o1[-5] * cs[ 5]) - (o2[ 5] * ca[ 5]);
			i2[ 5] = (o2[ 5] * cs[ 5]) + (o1[-5] * ca[ 5]);
			i1[-6] = (o1[-6] * cs[ 6]) - (o2[ 6] * ca[ 6]);
			i2[ 6] = (o2[ 6] * cs[ 6]) + (o1[-6] * ca[ 6]);
			i1[-7] = (o1[-7] * cs[ 7]) - (o2[ 7] * ca[ 7]);
			i2[ 7] = (o2[ 7] * cs[ 7]) + (o1[-7] * ca[ 7]);
			i1[-9] = o1[-9];
			i1[-8] = o1[-8];
      }
// ####
	memcpy( &hybridIn[31][8], &re_hybridOut[31][8], sizeof(real) * 10);
/*      hybridIn[31][8]  = re_hybridOut[31][8];
      hybridIn[31][9]  = re_hybridOut[31][9];
      hybridIn[31][10] = re_hybridOut[31][10];
      hybridIn[31][11] = re_hybridOut[31][11];
      hybridIn[31][12] = re_hybridOut[31][12];
      hybridIn[31][13] = re_hybridOut[31][13];
      hybridIn[31][14] = re_hybridOut[31][14];
      hybridIn[31][15] = re_hybridOut[31][15];
      hybridIn[31][16] = re_hybridOut[31][16];
      hybridIn[31][17] = re_hybridOut[31][17];
*/   }
}


void LayerIII_Decoder::inv_mdct(real *in, real *out, int32 block_type)
{
/*
This uses Byeong Gi Lee's Fast Cosine Transform algorithm to
decompose the 36 point and 12 point IDCT's into 9 point and 3
point IDCT's, respectively. Then the 9 point IDCT is computed
by a modified version of Mikko Tommila's IDCT algorithm, based on
the WFTA. See his comments before the first 9 point IDCT. The 3
point IDCT is already efficient to implement. -- Jeff Tsay. */
/* A little bit of improvement was done by me. - Hendrik */

/*------------------------------------------------------------------*/
/*                                                                  */
/*    Function: Calculation of the inverse MDCT                     */
/*    In the case of short blocks the 3 output vectors are already  */
/*    overlapped and added in this modul.                           */
/*                                                                  */
/*    New layer3                                                    */
/*                                                                  */
/*------------------------------------------------------------------*/

int32   i, six_i, p;
int32   odd_i, two_odd_i, four_odd_i, eight_odd_i;
real    tmp[18], save, sum;
real 	  pp1, pp2;
static  BOOL MDCT_init = FALSE;
static  real  win[4][36];

	 if(!MDCT_init){

	 /* type 0 */
		for(i=0;i<36;i++)
			win[0][i] = (real) sin( PI36 *(i+0.5) );

	 /* type 1*/
		for(i=0;i<18;i++)
			win[1][i] = (real) sin( PI36 *(i+0.5) );
		for(i=18;i<24;i++)
			win[1][i] = 1.0f;
		for(i=24;i<30;i++)
			win[1][i] = (real) sin( PI12 *(i+0.5-18) );
		for(i=30;i<36;i++)
			win[1][i] = 0.0f;

	 /* type 2 (not needed anymore) */
/*		for(i=0;i<12;i++)
			win[2][i] = (real) sin( PI12*(i+0.5) ) ;
		for(i=12;i<36;i++)
			win[2][i] = 0.0f; */

	 /* type 3*/
		for(i=0;i<6;i++)
			win[3][i] = 0.0f;
		for(i=6;i<12;i++)
			win[3][i] = (real) sin(PI12 * (i+ 0.5 - 6.0));
		for(i=12;i<18;i++)
			win[3][i] =1.0f;
		for(i=18;i<36;i++)
			win[3][i] = (real) sin(PI36 * (i + 0.5));

// #### saves twiddling factors for 36 point IDCT
		for(i=0;i<4;i++) {
			win[i][0] *= 0.740093616f;
			win[i][1] *= 0.821339815f;
			win[i][2] *= 0.930579498f;
			win[i][3] *= 1.082840285f;
			win[i][4] *= 1.306562965f;
			win[i][5] *= 1.662754762f;
			win[i][6] *= 2.310113158f;
			win[i][7] *= 3.830648788f;
			win[i][8] *= 11.46279281f;

			win[i][9] *= -11.46279281f;
			win[i][10] *= -3.830648788f;
			win[i][11] *= -2.310113158f;
			win[i][12] *= -1.662754762f;
			win[i][13] *= -1.306562965f;
			win[i][14] *= -1.082840285f;
			win[i][15] *= -0.930579498f;
			win[i][16] *= -0.821339815f;
			win[i][17] *= -0.740093616f;
			win[i][18] *= -0.678170852f;
			win[i][19] *= -0.630236207f;
			win[i][20] *= -0.592844523f;
			win[i][21] *= -0.563690973f;
			win[i][22] *= -0.541196100f;
			win[i][23] *= -0.524264562f;
			win[i][24] *= -0.512139757f;
			win[i][25] *= -0.504314480f;
			win[i][26] *= -0.500476342f;

			win[i][27] *= -0.500476342f;
			win[i][28] *= -0.504314480f;
			win[i][29] *= -0.512139757f;
			win[i][30] *= -0.524264562f;
			win[i][31] *= -0.541196100f;
			win[i][32] *= -0.563690973f;
			win[i][33] *= -0.592844523f;
			win[i][34] *= -0.630236207f;
			win[i][35] *= -0.678170852f;
		}

		MDCT_init = TRUE;
	 }

	 if(block_type == 2){
	   out[ 0] = out[ 1] = out[ 2] = out[ 3] = out[ 4] = out[ 5] = out[ 6] = out[ 7] = out[ 8] = out[ 9] =
	   out[10] = out[11] = out[12] = out[13] = out[14] = out[15] = out[16] = out[17] = out[18] = out[19] =
	   out[20] = out[21] = out[22] = out[23] = out[24] = out[25] = out[26] = out[27] = out[28] = out[29] =
	   out[30] = out[31] = out[32] = out[33] = out[34] = out[35] = 0.0f;
/*
	   for(p=0;p<36;p+=9) {
	   	out[p]   = out[p+1] = out[p+2] = out[p+3] =
	      out[p+4] = out[p+5] = out[p+6] = out[p+7] =
	      out[p+8] = 0.0f;
	   }
*/

   	six_i = 0;
	   for(i=0;i<3;i++)
   	{

      // 12 point IMDCT

      // Begin 12 point IDCT
// ####
	// Input aliasing for 12 pt IDCT
	// Input aliasing on odd indices (for 6 point IDCT)
	register real *inp = &in[i];
	inp[15]+=(inp[12]+=inp[ 9])+inp[6];
	inp[ 9]+=(inp[ 6]+=inp[ 3])+inp[0];
	inp[ 3]+= inp[ 0];
/* ####
	// Input aliasing for 12 pt IDCT
	   in[15+i] += in[12+i]; in[12+i] += in[9+i]; in[9+i]  +=  in[6+i];
   	in[6+i]  += in[3+i];  in[3+i]  += in[0+i];

	   // Input aliasing on odd indices (for 6 point IDCT)
   	in[15+i] += in[9+i];  in[9+i]  += in[3+i];
*/

	   // 3 point IDCT on even indices

{ // ####
	register real t1, t2;
	   tmp[1] = (t1 = inp[0]) - ( t2 = inp[12]);
	   sum = t1 + t2 * 0.500000000f;
	   pp1 = inp[ 6] * 0.866025403f;
	   tmp[0] = sum + pp1;
	   tmp[2] = sum - pp1;
}
      // End 3 point IDCT on even indices

	// 3 point IDCT on odd indices (for 6 point IDCT)

{ // ####
	register real t1, t2;
   	tmp[4] = ((t1 = inp[3]) - (t2 = inp[15])) * 0.707106781f;
   	sum = t1 + t2 * 0.500000000f;
   	pp1 = inp[9]  * 0.866025403f;
   	tmp[5] = (sum + pp1) * 0.517638090f;
   	tmp[3] = (sum - pp1) * 1.931851653f;
}

      // End 3 point IDCT on odd indices

   	// Twiddle factors on odd indices (for 6 point IDCT)
/*
   	tmp[3] *= 1.931851653f;
   	tmp[4] *= 0.707106781f;
   	tmp[5] *= 0.517638090f;
*/
   	// Output butterflies on 2 3 point IDCT's (for 6 point IDCT)

   	save = tmp[0];
   	tmp[0] += tmp[5];
   	tmp[5] = save - tmp[5];
	save = tmp[1];
	tmp[1] += tmp[4];
	tmp[4] = save - tmp[4];
	save = tmp[2];
	tmp[2] += tmp[3];
	tmp[3] = save - tmp[3];

   	// End 6 point IDCT

	   // Twiddle factors on indices (for 12 point IDCT)
/* // ####
	   tmp[0]  *=  0.504314480f;
	   tmp[1]  *=  0.541196100f;
	   tmp[2]  *=  0.630236207f;
	   tmp[3]  *=  0.821339815f;
	   tmp[4]  *=  1.306562965f;
	   tmp[5]  *=  3.830648788f;
*/
      // End 12 point IDCT
// ####
	   // Shift to 12 point modified IDCT, multiply by window type 2

	register real *outp = &out[six_i];
		outp[14] += tmp[0] * -0.793353340f * 0.504314480f;
		outp[15] += tmp[0] * -0.608761429f * 0.504314480f;
		outp[13] += tmp[1] * -0.923879532f * 0.541196100f;
		outp[16] += tmp[1] * -0.382683432f * 0.541196100f;
		outp[12] += tmp[2] * -0.991444861f * 0.630236207f;
		outp[17] += tmp[2] * -0.130526192f * 0.630236207f;

	register real t1 = tmp[3] * 0.130526192f * 0.821339815f;
		outp[ 6] += t1;
		outp[ 7] += tmp[4] * 0.382683432f * 1.306562965f;
		outp[ 8] += tmp[5] * 0.608761429f * 3.830648788f;

		outp[ 9] += tmp[5] * -0.793353340f * 3.830648788f;
		outp[10] += tmp[4] * -0.923879532f * 1.306562965f;
		outp[11] += t1 * (-0.991444861f/0.130526192f);

		six_i += 6;
   	}

	 } else {

   // 36 point IDCT

/* ####
   // input aliasing for 36 point IDCT
   in[17]+=in[16]; in[16]+=in[15]; in[15]+=in[14]; in[14]+=in[13];
   in[13]+=in[12]; in[12]+=in[11]; in[11]+=in[10]; in[10]+=in[9];
   in[9] +=in[8];  in[8] +=in[7];  in[7] +=in[6];  in[6] +=in[5];
   in[5] +=in[4];  in[4] +=in[3];  in[3] +=in[2];  in[2] +=in[1];
   in[1] +=in[0];

   // 18 point IDCT for odd indices

   // input aliasing for 18 point IDCT
   in[17]+=in[15]; in[15]+=in[13]; in[13]+=in[11]; in[11]+=in[9];
   in[9] +=in[7];  in[7] +=in[5];  in[5] +=in[3];  in[3] +=in[1];
*/
// #### combined the aliasing for fewer storage ops
   in[17]+=(in[16]+=in[15])+in[14]; in[15]+=(in[14]+=in[13])+in[12];
   in[13]+=(in[12]+=in[11])+in[10]; in[11]+=(in[10]+=in[ 9])+in[ 8];
   in[ 9]+=(in[ 8]+=in[ 7])+in[ 6]; in[ 7]+=(in[ 6]+=in[ 5])+in[ 4];
   in[ 5]+=(in[ 4]+=in[ 3])+in[ 2]; in[ 3]+=(in[ 2]+=in[ 1])+in[ 0];
   in[ 1]+=in[ 0];

	   // 9 point IDCT on even indices

		// original:

/*   for(i=0; i<9; i++) {
   	sum = 0.0;

      for(j=0;j<18;j+=2)
      	sum += in[j] * cos(PI36 * (2*i + 1) * j);

      tmp[i] = sum;
     } */

// 9 Point Inverse Discrete Cosine Transform
//
// This piece of code is Copyright 1997 Mikko Tommila and is freely usable
// by anybody. The algorithm itself is of course in the public domain.
//
// Again derived heuristically from the 9-point WFTA.
//
// The algorithm is optimized (?) for speed, not for small rounding errors or
// good readability.
//
// 36 additions, 11 multiplications
//
// Again this is very likely sub-optimal.
//
// The code is optimized to use a minimum number of temporary variables,
// so it should compile quite well even on 8-register Intel x86 processors.
// This makes the code quite obfuscated and very difficult to understand.
//
// References:
// [1] S. Winograd: "On Computing the Discrete Fourier Transform",
//     Mathematics of Computation, Volume 32, Number 141, January 1978,
//     Pages 175-199

// Some modifications for maplay by Jeff Tsay
// #### further modified by Hendrik Pagenhardt to reduce memory accesses and 2 adds
      {
	real t0 ,t1, t2, t3, t4, t5, t6, t7, t8, t9 ,t10, t11, t12, t13, t14, t15, t16, t17;

	t0 = in[0] + COSPI3 * in[12];
	t1 = in[0] - in[12];
	t2 = in[8] + in[16] - in[4];

    	tmp[4] = t1 + t2;

    	t3 = t1 - COSPI3 * t2;
    	t4 = COSPI6 * (in[10] + in[14] - in[2]);

    	tmp[1] = t3 - t4;
    	tmp[7] = t3 + t4;

    	t5 = DCTEVEN1 * (in[4] + in[8]);
    	t6 = DCTEVEN2 * (in[8] - in[16]);
    	t7 = DCTEVEN3 * (in[4] + in[16]);

    	t17= t0 - t5 - t7;
    	t8 = t5 + t0 + t6;
    	t9 = t0 + t7 - t6;

    	t12 = COSPI6 * in[6];
    	t10 = DCTODD1 * (in[2]  + in[10]);
    	t11 = DCTODD2 * (in[10] - in[14]);

    	t13 = t10 + t11 + t12;

    	tmp[0] = t8 + t13;
    	tmp[8] = t8 - t13;

    	t14 = DCTODD3 * (in[2] + in[14]);
    	t15 = t10 + t14 - t12;
    	t16 = t11 - t14 - t12;

    	tmp[3] = t9 + t15;
    	tmp[5] = t9 - t15;

    	tmp[2] = t17 + t16;
    	tmp[6] = t17 - t16;
     }
   	// End 9 point IDCT on even indices
   // 9 point IDCT on odd indices

	// original:
/*   for(i=0; i<9; i++) {
   	sum = 0.0;

      for(j=0;j<18;j+=2)
      	sum += in[j+1] * cos(PI36 * (2*i + 1) * j);

      tmp[17-i] = sum;
     } */
   	// This includes multiplication by the twiddle factors
      // at the end -- Jeff.
     {
	real t0 ,t1, t2, t3, t4, t5, t6, t7, t8, t9 ,t10, t11, t12, t13, t14, t15, t16, t17;

	t0 = in[1] + COSPI3 * in[13];
	t1 = in[1] - in[13];
	t2 = in[9] + in[17] - in[5];

    	tmp[13] = (t1 + t2) * 0.707106781f;

    	t3 = t1 - COSPI3 * t2;
    	t4 = COSPI6 * (in[11] + in[15] - in[3]);

    	tmp[16] = (t3 - t4) * 0.517638090f;
    	tmp[10] = (t3 + t4) * 1.931851653f;

    	t5 = DCTEVEN1 * (in[5] + in[9]);
    	t6 = DCTEVEN2 * (in[9] - in[17]);
    	t7 = DCTEVEN3 * (in[5] + in[17]);

    	t17= t0 - t5 - t7;
    	t8 = t5 + t0 + t6;
    	t9 = t0 + t7 - t6;

    	t12 = COSPI6  * in[7];
    	t10 = DCTODD1 * (in[3]  + in[11]);
    	t11 = DCTODD2 * (in[11] - in[15]);

    	t13 = t10 + t11 + t12;

    	tmp[17] = (t8 + t13) * 0.501909918f;
    	tmp[9]  = (t8 - t13) * 5.736856623f;

    	t14 = DCTODD3 * (in[3] + in[15]);
    	t15 = t10 + t14 - t12;
    	t16 = t11 - t14 - t12;

    	tmp[14] = (t9 + t15) * 0.610387294f;
    	tmp[12] = (t9 - t15) * 0.871723397f;

    	tmp[15] = (t17 + t16) * 0.551688959f;
    	tmp[11] = (t17 - t16) * 1.183100792f;
    	}

	   // End 9 point IDCT on odd indices

   // Butterflies on 9 point IDCT's
/*
   for (i=0;i<9;i++) {
      save = tmp[i];
      tmp[i] += tmp[17-i];
      tmp[17-i] = save - tmp[17-i];
   }
*/

// ####
real *tt[2] = {tmp,tmp};
register real *tmp1 = tt[0];
{      register real t1 = tmp1[0], t2 = tmp1[17-0]; tmp1[0] = t1 + t2; tmp1[17-0] = t1 - t2;}
{      register real t1 = tmp1[1], t2 = tmp1[17-1]; tmp1[1] = t1 + t2; tmp1[17-1] = t1 - t2;}
{      register real t1 = tmp1[2], t2 = tmp1[17-2]; tmp1[2] = t1 + t2; tmp1[17-2] = t1 - t2;}
{      register real t1 = tmp1[3], t2 = tmp1[17-3]; tmp1[3] = t1 + t2; tmp1[17-3] = t1 - t2;}
{      register real t1 = tmp1[4], t2 = tmp1[17-4]; tmp1[4] = t1 + t2; tmp1[17-4] = t1 - t2;}
{      register real t1 = tmp1[5], t2 = tmp1[17-5]; tmp1[5] = t1 + t2; tmp1[17-5] = t1 - t2;}
{      register real t1 = tmp1[6], t2 = tmp1[17-6]; tmp1[6] = t1 + t2; tmp1[17-6] = t1 - t2;}
{      register real t1 = tmp1[7], t2 = tmp1[17-7]; tmp1[7] = t1 + t2; tmp1[17-7] = t1 - t2;}
{      register real t1 = tmp1[8], t2 = tmp1[17-8]; tmp1[8] = t1 + t2; tmp1[17-8] = t1 - t2;}
/*
   real *tmp1 = tmp, *tmp2 = &tmp[17];
   i = 9;
   do {
      save = *tmp1;
      *tmp1++ += *tmp2;
      *tmp2-- = save - *tmp2;
   } while (--i);
*/
   // end 18 point IDCT

   // twiddle factors for 36 point IDCT
/* // #### not needed, done for win[i]
   tmp[0] *=  -0.500476342f;
   tmp[1] *=  -0.504314480f;
   tmp[2] *=  -0.512139757f;
   tmp[3] *=  -0.524264562f;
   tmp[4] *=  -0.541196100f;
   tmp[5] *=  -0.563690973f;
   tmp[6] *=  -0.592844523f;
   tmp[7] *=  -0.630236207f;
   tmp[8] *=  -0.678170852f;
   tmp[9] *=  -0.740093616f;
   tmp[10]*=  -0.821339815f;
   tmp[11]*=  -0.930579498f;
   tmp[12]*=  -1.082840285f;
   tmp[13]*=  -1.306562965f;
   tmp[14]*=  -1.662754762f;
   tmp[15]*=  -2.310113158f;
   tmp[16]*=  -3.830648788f;
   tmp[17]*= -11.46279281f;
*/
   // end 36 point IDCT

   // shift to modified IDCT

   register float *winp = win[block_type];
   out[0] = tmp[9]  * winp[0]; // #### "-" done for winp
   out[1] = tmp[10] * winp[1]; // #### "-" done for winp
   out[2] = tmp[11] * winp[2]; // #### "-" done for winp
   out[3] = tmp[12] * winp[3]; // #### "-" done for winp
   out[4] = tmp[13] * winp[4]; // #### "-" done for winp
   out[5] = tmp[14] * winp[5]; // #### "-" done for winp
   out[6] = tmp[15] * winp[6]; // #### "-" done for winp
   out[7] = tmp[16] * winp[7]; // #### "-" done for winp
   out[8] = tmp[17] * winp[8]; // #### "-" done for winp
   out[9] = tmp[17] * winp[9];
   out[10]= tmp[16] * winp[10];
   out[11]= tmp[15] * winp[11];
   out[12]= tmp[14] * winp[12];
   out[13]= tmp[13] * winp[13];
   out[14]= tmp[12] * winp[14];
   out[15]= tmp[11] * winp[15];
   out[16]= tmp[10] * winp[16];
   out[17]= tmp[9]  * winp[17];
   out[18]= tmp[8]  * winp[18];
   out[19]= tmp[7]  * winp[19];
   out[20]= tmp[6]  * winp[20];
   out[21]= tmp[5]  * winp[21];
   out[22]= tmp[4]  * winp[22];
   out[23]= tmp[3]  * winp[23];
   out[24]= tmp[2]  * winp[24];
   out[25]= tmp[1]  * winp[25];
   out[26]= tmp[0]  * winp[26];
   out[27]= tmp[0]  * winp[27];
   out[28]= tmp[1]  * winp[28];
   out[29]= tmp[2]  * winp[29];
   out[30]= tmp[3]  * winp[30];
   out[31]= tmp[4]  * winp[31];
   out[32]= tmp[5]  * winp[32];
   out[33]= tmp[6]  * winp[33];
   out[34]= tmp[7]  * winp[34];
   out[35]= tmp[8]  * winp[35];
   }
}


void LayerIII_Decoder::decode()
{
  int32 nSlots = header->slots();
  int32 gr, ch, main_data_end, flush_main;
  int32 bytes_to_discard;

  stream->get_side_info(channels, si);

  for (; nSlots > 0; nSlots--)  // read main data.
    br->hputbuf((uint32) stream->get_bits(8));

  main_data_end = br->hsstell() >> 3; // of previous frame

  if (flush_main=(br->hsstell() & 0x7)) {
    br->hgetbits(8 - flush_main);
    main_data_end++;
  }

  bytes_to_discard = frame_start - main_data_end - si->main_data_begin;
  if(main_data_end > 4096) {
    frame_start -= 4096;
    br->rewindNbytes(4096);
  }

  frame_start += header->slots();

  if (bytes_to_discard < 0)
    return;

// ####
  if (bytes_to_discard > 0)
	br->rewindNbytes(- bytes_to_discard);
//  for (; bytes_to_discard > 0; bytes_to_discard--) br->hgetbits(8);

  for (gr=0;gr<max_gr;gr++) {

    for (ch=0; ch<channels; ch++) {
      part2_start= br->hsstell();
      get_scale_factors(ch, gr);
      hufman_decode(ch, gr);
      dequantize_sample(ro[ch], ch, gr);
    }

    stereo(gr);

    for (ch=first_channel; ch<=last_channel; ch++) { 

      reorder (lr[ch], ch, gr);
      antialias(ch, gr);

      for (register int32 sb=0;sb<SBLIMIT;sb++) { // Hybrid synthesis.
	hybrid(ch, gr, sb);
      }
/* #### done in hybrid()
      for (register int32 ss=1;ss<SSLIMIT;ss+=2) // Frequency inversion for polyphase.
	for (register int32 sb=1;sb<SBLIMIT;sb+=2)
	  re_hybridOut[sb][ss] = -re_hybridOut[sb][ss];
*/
      if ((ch == 0) || (which_channels == right))
// ####
	for (register int32 ss=0;ss<SSLIMIT;ss++) { // Polyphase synthesis
	    register real *rep = &re_hybridOut[0][ss], *sp = filter1->samples;
	    sp[ 0] = rep[ 0 * SSLIMIT];
	    sp[ 1] = rep[ 1 * SSLIMIT];
	    sp[ 2] = rep[ 2 * SSLIMIT];
	    sp[ 3] = rep[ 3 * SSLIMIT];
	    sp[ 4] = rep[ 4 * SSLIMIT];
	    sp[ 5] = rep[ 5 * SSLIMIT];
	    sp[ 6] = rep[ 6 * SSLIMIT];
	    sp[ 7] = rep[ 7 * SSLIMIT];
	    sp[ 8] = rep[ 8 * SSLIMIT];
	    sp[ 9] = rep[ 9 * SSLIMIT];
	    sp[10] = rep[10 * SSLIMIT];
	    sp[11] = rep[11 * SSLIMIT];
	    sp[12] = rep[12 * SSLIMIT];
	    sp[13] = rep[13 * SSLIMIT];
	    sp[14] = rep[14 * SSLIMIT];
	    sp[15] = rep[15 * SSLIMIT];
	    sp[16] = rep[16 * SSLIMIT];
	    sp[17] = rep[17 * SSLIMIT];
	    sp[18] = rep[18 * SSLIMIT];
	    sp[19] = rep[19 * SSLIMIT];
	    sp[20] = rep[20 * SSLIMIT];
	    sp[21] = rep[21 * SSLIMIT];
	    sp[22] = rep[22 * SSLIMIT];
	    sp[23] = rep[23 * SSLIMIT];
	    sp[24] = rep[24 * SSLIMIT];
	    sp[25] = rep[25 * SSLIMIT];
	    sp[26] = rep[26 * SSLIMIT];
	    sp[27] = rep[27 * SSLIMIT];
	    sp[28] = rep[28 * SSLIMIT];
	    sp[29] = rep[29 * SSLIMIT];
	    sp[30] = rep[30 * SSLIMIT];
	    sp[31] = rep[31 * SSLIMIT];

	  filter1->calculate_pcm_samples(buffer);
	}
/*
	for (register int32 ss=0;ss<SSLIMIT;ss++) { // Polyphase synthesis
	  for (register int32 sb=0;sb<SBLIMIT;sb++)
	    filter1->input_sample(re_hybridOut[sb][ss], sb);

	  filter1->calculate_pcm_samples(buffer);
	}
*/
      else
	for (register int32 ss=0;ss<SSLIMIT;ss++) { // Polyphase synthesis
	    register real *rep = &re_hybridOut[0][ss], *sp = filter2->samples;
	    sp[ 0] = rep[ 0 * SSLIMIT];
	    sp[ 1] = rep[ 1 * SSLIMIT];
	    sp[ 2] = rep[ 2 * SSLIMIT];
	    sp[ 3] = rep[ 3 * SSLIMIT];
	    sp[ 4] = rep[ 4 * SSLIMIT];
	    sp[ 5] = rep[ 5 * SSLIMIT];
	    sp[ 6] = rep[ 6 * SSLIMIT];
	    sp[ 7] = rep[ 7 * SSLIMIT];
	    sp[ 8] = rep[ 8 * SSLIMIT];
	    sp[ 9] = rep[ 9 * SSLIMIT];
	    sp[10] = rep[10 * SSLIMIT];
	    sp[11] = rep[11 * SSLIMIT];
	    sp[12] = rep[12 * SSLIMIT];
	    sp[13] = rep[13 * SSLIMIT];
	    sp[14] = rep[14 * SSLIMIT];
	    sp[15] = rep[15 * SSLIMIT];
	    sp[16] = rep[16 * SSLIMIT];
	    sp[17] = rep[17 * SSLIMIT];
	    sp[18] = rep[18 * SSLIMIT];
	    sp[19] = rep[19 * SSLIMIT];
	    sp[20] = rep[20 * SSLIMIT];
	    sp[21] = rep[21 * SSLIMIT];
	    sp[22] = rep[22 * SSLIMIT];
	    sp[23] = rep[23 * SSLIMIT];
	    sp[24] = rep[24 * SSLIMIT];
	    sp[25] = rep[25 * SSLIMIT];
	    sp[26] = rep[26 * SSLIMIT];
	    sp[27] = rep[27 * SSLIMIT];
	    sp[28] = rep[28 * SSLIMIT];
	    sp[29] = rep[29 * SSLIMIT];
	    sp[30] = rep[30 * SSLIMIT];
	    sp[31] = rep[31 * SSLIMIT];

	  filter2->calculate_pcm_samples(buffer);
	}
/*
	for (register int32 ss=0;ss<SSLIMIT;ss++) { // Polyphase synthesis
	  for (register int32 sb=0;sb<SBLIMIT;sb++)
	    filter2->input_sample(re_hybridOut[sb][ss], sb);

	  filter2->calculate_pcm_samples(buffer);
	}
*/
    }	// channels
  }	// granule
  buffer->write_buffer (1);		// write to stdout
}

