/* Layer III decoder object declarations. Rearrangement
   of layer III routines as an object by Jeff Tsay.

   Last modified : 05/07/97 */

#ifndef LAYERIII_H
#define LAYERIII_H

#include "all.h"
#include "l3type.h"
#include "ibitstream.h"
#include "obuffer.h"
#include "bit_res.h"
#include "header.h"
#include "synthesis_filter.h"

#define SSLIMIT 18
#define SBLIMIT 32

class LayerIII_Decoder {

public:

  LayerIII_Decoder(char *huffdec_path0, Ibitstream *stream0, Header *header0,
		   SynthesisFilter *filtera, SynthesisFilter *filterb,
		   Obuffer *buffer0, e_channels);

  ~LayerIII_Decoder();

  // Notify decoder that a seek is being made
  void seek_notify();

  // Decode one frame, filling the buffer with the output samples
  void decode();


private:

  // private data

  int32 is[SBLIMIT][SSLIMIT];
  real  ro[2][SBLIMIT][SSLIMIT];
  real  lr[2][SBLIMIT][SSLIMIT];
  real  hybridIn[SBLIMIT][SSLIMIT];
  real  re_hybridOut[SBLIMIT][SSLIMIT];
  real prevblck[2][SBLIMIT][SSLIMIT];

  char *huffdec_path;
  Ibitstream *stream;
  Header *header;
  SynthesisFilter *filter1, *filter2;
  Obuffer *buffer;
  enum e_channels which_channels;
  Bit_Reserve *br;
  III_side_info_t *si;
  III_scalefac_t scalefac;

  uint32 max_gr;
  int32 frame_start;
  int32 part2_start;
  uint32 channels;
  uint32 first_channel;
  uint32 last_channel;

  // private member functions

  void get_scale_factors(int32 ch, int32 gr);
  void hufman_decode(int32 ch, int32 gr);
  real two_pow2(int32 scale, int32 preflag, int32 pretab_offset, int32 l);
  void dequantize_sample(real xr[SBLIMIT][SSLIMIT], int32 ch, int32 gr);
  void reorder(real xr[SBLIMIT][SSLIMIT], int32 ch, int32 gr);
  void stereo(int32 gr);
  void antialias(int32 ch, int32 gr);
  void inv_mdct(real *in, real *out, int32 block_type);
  void do_downmix();
//  void hybrid(int32 ch, int32 gr, int32 sb);
inline void LayerIII_Decoder::hybrid(int32 ch, int32 gr, int32 sb)
{

	real *fsIn  = hybridIn[sb];
   real *tsOut = re_hybridOut[sb];
   gr_info_s *gr_info = &(si->ch[ch].gr[gr]);
// ####
	real *pb = prevblck[ch][sb];

	real rawout[36];
	int32 bt;

	bt = (gr_info->window_switching_flag && gr_info->mixed_block_flag &&
			 (sb < 2)) ? 0 : gr_info->block_type;

	inv_mdct(fsIn, rawout, bt);

	/* overlap addition */
// ####
if (sb & 1) {
   tsOut[0]             = rawout[0]  + pb[0];
   pb[0]  = rawout[18];
   tsOut[1]             = -(rawout[1]  + pb[1]);
   pb[1]  = rawout[19];
   tsOut[2]             = rawout[2]  + pb[2];
   pb[2]  = rawout[20];
   tsOut[3]             = -(rawout[3]  + pb[3]);
   pb[3]  = rawout[21];
   tsOut[4]             = rawout[4]  + pb[4];
   pb[4]  = rawout[22];
   tsOut[5]             = -(rawout[5]  + pb[5]);
   pb[5]  = rawout[23];
   tsOut[6]             = rawout[6]  + pb[6];
   pb[6]  = rawout[24];
   tsOut[7]             = -(rawout[7]  + pb[7]);
   pb[7]  = rawout[25];
   tsOut[8]             = rawout[8]  + pb[8];
   pb[8]  = rawout[26];
   tsOut[9]             = -(rawout[9]  + pb[9]);
   pb[9]  = rawout[27];
   tsOut[10]            = rawout[10] + pb[10];
   pb[10] = rawout[28];
   tsOut[11]            = -(rawout[11] + pb[11]);
   pb[11] = rawout[29];
   tsOut[12]            = rawout[12] + pb[12];
   pb[12] = rawout[30];
   tsOut[13]            = -(rawout[13] + pb[13]);
   pb[13] = rawout[31];
   tsOut[14]            = rawout[14] + pb[14];
   pb[14] = rawout[32];
   tsOut[15]            = -(rawout[15] + pb[15]);
   pb[15] = rawout[33];
   tsOut[16]            = rawout[16] + pb[16];
   pb[16] = rawout[34];
   tsOut[17]            = -(rawout[17] + pb[17]);
   pb[17] = rawout[35];

/*   tsOut[0]             = rawout[0]  + prevblck[ch][sb][0];
   prevblck[ch][sb][0]  = rawout[18];
   tsOut[1]             = -(rawout[1]  + prevblck[ch][sb][1]);
   prevblck[ch][sb][1]  = rawout[19];
   tsOut[2]             = rawout[2]  + prevblck[ch][sb][2];
   prevblck[ch][sb][2]  = rawout[20];
   tsOut[3]             = -(rawout[3]  + prevblck[ch][sb][3]);
   prevblck[ch][sb][3]  = rawout[21];
   tsOut[4]             = rawout[4]  + prevblck[ch][sb][4];
   prevblck[ch][sb][4]  = rawout[22];
   tsOut[5]             = -(rawout[5]  + prevblck[ch][sb][5]);
   prevblck[ch][sb][5]  = rawout[23];
   tsOut[6]             = rawout[6]  + prevblck[ch][sb][6];
   prevblck[ch][sb][6]  = rawout[24];
   tsOut[7]             = -(rawout[7]  + prevblck[ch][sb][7]);
   prevblck[ch][sb][7]  = rawout[25];
   tsOut[8]             = rawout[8]  + prevblck[ch][sb][8];
   prevblck[ch][sb][8]  = rawout[26];
   tsOut[9]             = -(rawout[9]  + prevblck[ch][sb][9]);
   prevblck[ch][sb][9]  = rawout[27];
   tsOut[10]            = rawout[10] + prevblck[ch][sb][10];
   prevblck[ch][sb][10] = rawout[28];
   tsOut[11]            = -(rawout[11] + prevblck[ch][sb][11]);
   prevblck[ch][sb][11] = rawout[29];
   tsOut[12]            = rawout[12] + prevblck[ch][sb][12];
   prevblck[ch][sb][12] = rawout[30];
   tsOut[13]            = -(rawout[13] + prevblck[ch][sb][13]);
   prevblck[ch][sb][13] = rawout[31];
   tsOut[14]            = rawout[14] + prevblck[ch][sb][14];
   prevblck[ch][sb][14] = rawout[32];
   tsOut[15]            = -(rawout[15] + prevblck[ch][sb][15]);
   prevblck[ch][sb][15] = rawout[33];
   tsOut[16]            = rawout[16] + prevblck[ch][sb][16];
   prevblck[ch][sb][16] = rawout[34];
   tsOut[17]            = -(rawout[17] + prevblck[ch][sb][17]);
   prevblck[ch][sb][17] = rawout[35];
*/
} else {
   tsOut[0]             = rawout[0]  + pb[0];
   pb[0]  = rawout[18];
   tsOut[1]             = rawout[1]  + pb[1];
   pb[1]  = rawout[19];
   tsOut[2]             = rawout[2]  + pb[2];
   pb[2]  = rawout[20];
   tsOut[3]             = rawout[3]  + pb[3];
   pb[3]  = rawout[21];
   tsOut[4]             = rawout[4]  + pb[4];
   pb[4]  = rawout[22];
   tsOut[5]             = rawout[5]  + pb[5];
   pb[5]  = rawout[23];
   tsOut[6]             = rawout[6]  + pb[6];
   pb[6]  = rawout[24];
   tsOut[7]             = rawout[7]  + pb[7];
   pb[7]  = rawout[25];
   tsOut[8]             = rawout[8]  + pb[8];
   pb[8]  = rawout[26];
   tsOut[9]             = rawout[9]  + pb[9];
   pb[9]  = rawout[27];
   tsOut[10]            = rawout[10] + pb[10];
   pb[10] = rawout[28];
   tsOut[11]            = rawout[11] + pb[11];
   pb[11] = rawout[29];
   tsOut[12]            = rawout[12] + pb[12];
   pb[12] = rawout[30];
   tsOut[13]            = rawout[13] + pb[13];
   pb[13] = rawout[31];
   tsOut[14]            = rawout[14] + pb[14];
   pb[14] = rawout[32];
   tsOut[15]            = rawout[15] + pb[15];
   pb[15] = rawout[33];
   tsOut[16]            = rawout[16] + pb[16];
   pb[16] = rawout[34];
   tsOut[17]            = rawout[17] + pb[17];
   pb[17] = rawout[35];
/*
   tsOut[0]             = rawout[0]  + prevblck[ch][sb][0];
   prevblck[ch][sb][0]  = rawout[18];
   tsOut[1]             = rawout[1]  + prevblck[ch][sb][1];
   prevblck[ch][sb][1]  = rawout[19];
   tsOut[2]             = rawout[2]  + prevblck[ch][sb][2];
   prevblck[ch][sb][2]  = rawout[20];
   tsOut[3]             = rawout[3]  + prevblck[ch][sb][3];
   prevblck[ch][sb][3]  = rawout[21];
   tsOut[4]             = rawout[4]  + prevblck[ch][sb][4];
   prevblck[ch][sb][4]  = rawout[22];
   tsOut[5]             = rawout[5]  + prevblck[ch][sb][5];
   prevblck[ch][sb][5]  = rawout[23];
   tsOut[6]             = rawout[6]  + prevblck[ch][sb][6];
   prevblck[ch][sb][6]  = rawout[24];
   tsOut[7]             = rawout[7]  + prevblck[ch][sb][7];
   prevblck[ch][sb][7]  = rawout[25];
   tsOut[8]             = rawout[8]  + prevblck[ch][sb][8];
   prevblck[ch][sb][8]  = rawout[26];
   tsOut[9]             = rawout[9]  + prevblck[ch][sb][9];
   prevblck[ch][sb][9]  = rawout[27];
   tsOut[10]            = rawout[10] + prevblck[ch][sb][10];
   prevblck[ch][sb][10] = rawout[28];
   tsOut[11]            = rawout[11] + prevblck[ch][sb][11];
   prevblck[ch][sb][11] = rawout[29];
   tsOut[12]            = rawout[12] + prevblck[ch][sb][12];
   prevblck[ch][sb][12] = rawout[30];
   tsOut[13]            = rawout[13] + prevblck[ch][sb][13];
   prevblck[ch][sb][13] = rawout[31];
   tsOut[14]            = rawout[14] + prevblck[ch][sb][14];
   prevblck[ch][sb][14] = rawout[32];
   tsOut[15]            = rawout[15] + prevblck[ch][sb][15];
   prevblck[ch][sb][15] = rawout[33];
   tsOut[16]            = rawout[16] + prevblck[ch][sb][16];
   prevblck[ch][sb][16] = rawout[34];
   tsOut[17]            = rawout[17] + prevblck[ch][sb][17];
   prevblck[ch][sb][17] = rawout[35];
*/
}
}


};

#endif
