/*
 * DART/MMPM playback is now possible through different Obuffer classes
 * (selected in maplay.cc)
 * DART routines by ... Darwin O' ...
 * Hendrik Pagenhardt (pagenhar@sunpool.cs.uni-magdeburg.de)
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>
#include "obuffer.h"
#include "header.h"

FileObuffer::FileObuffer (uint32 number_of_channels)
{
#ifdef DEBUG
  if (!number_of_channels || number_of_channels > MAXCHANNELS)
  {
    fprintf(stderr,"FileObuffer: number of channels has to be in [1,%d] !\n",MAXCHANNELS);
    exit (1);
  }
#endif

  fop=fopen("testout.raw","wb");

  channels = number_of_channels;
  for (int i = 0; i < number_of_channels; ++i)
    bufferp[i] = buffer + i;

}


void FileObuffer::append (uint32 channel, int16 value)
{
#ifdef DEBUG
  if (channel >= channels)
  {
    fprintf(stderr,"illegal channelnumber in FileObuffer::append()!\n");
    exit (1);
  }
  if (bufferp[channel] - buffer >= OBUFFERSIZE)
  {
    fprintf(stderr,"FileObuffer: buffer overflow!\n");
    exit (1);
  }
#endif

  *bufferp[channel] = value;
  bufferp[channel] += channels;
}


void FileObuffer::write_buffer (int fd)
{
  int length = (int)((char *)bufferp[0] - (char *)buffer), writelength;

  writelength=fwrite ((char *)buffer,1, length,fop);	// #### comment this line out and start 
							// maplay with -s parameter to get raw
							// sound rendering speed. Hendrik

  for (int i = 0; i < channels; ++i)
    bufferp[i] = buffer + i;
}

#ifdef OS2
#define INCL_DOS
#define INCL_DOSPROCESS
#define INCL_NOPMAPI
#define INCL_OS2MM

#include <os2.h>
#include <os2me.h>

// the DART Obuffer
#define BUFMULT 12
#define PLSIZE 6

LONG MyEvent (ULONG ulStatus, PMCI_MIX_BUFFER pBuffer, ULONG ulFlags);

MCI_MIX_BUFFER Buffers[PLSIZE];
int numWritten = 0;
int buffring = 0;
HEV sem;
MCI_AMP_OPEN_PARMS AmpOpenParms = {NULLHANDLE,0,0,(PSZ)MCI_DEVTYPE_AUDIO_AMPMIX,NULL,NULL,NULL};
MCI_MIXSETUP_PARMS MixSetupParms = {NULLHANDLE,16,MCI_WAVE_FORMAT_PCM,0,0,MCI_PLAY,MCI_DEVTYPE_WAVEFORM_AUDIO,0,NULL,NULL,MyEvent,NULL,0,0};
MCI_BUFFER_PARMS BufferParms = {NULLHANDLE,sizeof(MCI_BUFFER_PARMS),PLSIZE,OBUFFERSIZE*2*BUFMULT,0,0,0,Buffers};

void error(ULONG rc) {
   unsigned char errstr[80];

   mciGetErrorString(rc,errstr,80);
   fprintf(stderr,"MMOS/2 error: %s", errstr);
}

LONG MyEvent (ULONG ulStatus, PMCI_MIX_BUFFER pBuffer, ULONG ulFlags) {
   if (ulFlags & MIX_WRITE_COMPLETE) {
	numWritten--;
	DosPostEventSem(sem);
   }
   return 0;
}

OS2ObufferDART::OS2ObufferDART (uint32 number_of_channels, Header *header) {
   ULONG rc;

  channels = number_of_channels;
   rc = mciSendCommand(0,MCI_OPEN,MCI_WAIT | MCI_OPEN_TYPE_ID,&AmpOpenParms,0);
   if (ULONG_LOWD(rc)!=MCIERR_SUCCESS) error(rc);
   MixSetupParms.ulSamplesPerSec=header->frequency ();
   MixSetupParms.ulChannels=channels;
   rc=mciSendCommand(AmpOpenParms.usDeviceID,MCI_MIXSETUP,
		     MCI_WAIT | MCI_MIXSETUP_INIT,&MixSetupParms,0);
   if (ULONG_LOWD(rc)!=MCIERR_SUCCESS) error(rc);
   rc=mciSendCommand(AmpOpenParms.usDeviceID,MCI_BUFFER,
                     MCI_WAIT | MCI_ALLOCATE_MEMORY,&BufferParms,0);
   if (ULONG_LOWD(rc)!=MCIERR_SUCCESS) error(rc);
   for (unsigned int i = 0; i < number_of_channels; ++i)
      bufferp[i] = (int16 *)Buffers[0].pBuffer + i;
   DosCreateEventSem(NULL,&sem,0,TRUE);
}   

void OS2ObufferDART::write_buffer (int) {
   static int count = 0;

   if ((count += channels) < BUFMULT * 2) return;
   count = 0;
   if (numWritten++ >= PLSIZE - 1) {
	ULONG numPosted;
	DosResetEventSem(sem, &numPosted);
	DosWaitEventSem(sem, (ULONG)SEM_INDEFINITE_WAIT);
   }
   Buffers[buffring].ulBufferLength =
      (int)((char *)bufferp[0] - (char *)Buffers[buffring].pBuffer);
   if (Buffers[buffring].ulBufferLength > OBUFFERSIZE * 2 * BUFMULT) {
      exit(1);
   }
   MixSetupParms.pmixWrite(MixSetupParms.ulMixHandle,&Buffers[buffring++],1);
   if (buffring >= PLSIZE) buffring = 0;
   for (unsigned int i = 0; i < channels; ++i)
      bufferp[i] = (int16 *)Buffers[buffring].pBuffer + i;
}
   
OS2ObufferDART::~OS2ObufferDART (void) {
   MCI_GENERIC_PARMS GenericParms;
   ULONG rc;

   while (numWritten > 0) {
	ULONG numPosted;
	DosResetEventSem(sem, &numPosted);
	DosWaitEventSem(sem, (ULONG)SEM_INDEFINITE_WAIT);
   }
   DosCloseEventSem(sem);
   rc=mciSendCommand(AmpOpenParms.usDeviceID,MCI_BUFFER,
                     MCI_WAIT | MCI_DEALLOCATE_MEMORY,&BufferParms,0);
   if (ULONG_LOWD(rc)!=MCIERR_SUCCESS) error(rc);
   rc=mciSendCommand(AmpOpenParms.usDeviceID,MCI_CLOSE,MCI_WAIT,&GenericParms,0);
   if (ULONG_LOWD(rc)!=MCIERR_SUCCESS) error(rc);
}


#ifdef DEBUG
void OS2ObufferDART::append (uint32 channel, int16 value)
{
  if (channel >= channels)
  {
    fprintf(stderr,"illegal channelnumber in OS2Obuffer::append()!\n");
    exit (1);
  }
  if (bufferp[channel] - buffer >= OBUFFERSIZE)
  {
    fprintf(stderr,"buffer overflow!\n");
    exit (1);
  }
  *bufferp[channel] = value;
  bufferp[channel] += channels;
}
#endif

// #### the MMPM Obuffer
#undef PLSIZE
#undef BUFMULT
#include "playwav.h"

#define BUFMULT 8

OS2ObufferMMPM::OS2ObufferMMPM (uint32 number_of_channels, Header *header)
{
  channels = number_of_channels;

  for (int i = 0; i < number_of_channels; ++i)
    bufferp[i] = buffer + i;

  // configure the device:
  int play_sample_rate = header->frequency ();
  p=new playwav(OBUFFERSIZE*2*BUFMULT,play_sample_rate,channels);
  bpos=0;next=bigbuf;
  memset(buffer,0,sizeof(buffer));
}


OS2ObufferMMPM::~OS2ObufferMMPM (void)
{
  int q;

  memset(buffer,0,sizeof(buffer));
  for(q=0;q<BUFMULT*PLSIZE*(3-channels)+1;q++) {
    bufferp[0]=&buffer[OBUFFERSIZE/(3-channels)];
    write_buffer(0);
  }
  delete p;
}


#ifdef DEBUG
void OS2ObufferMMPM::append (uint32 channel, int16 value)
{
  if (channel >= channels)
  {
    fprintf(stderr,"illegal channelnumber in OS2Obuffer::append()!\n");
    exit (1);
  }
  if (bufferp[channel] - buffer >= OBUFFERSIZE)
  {
    fprintf(stderr,"buffer overflow!\n");
    exit (1);
  }
  *bufferp[channel] = value;
  bufferp[channel] += channels;
}
#endif


void OS2ObufferMMPM::write_buffer (int)
{
  int length = (int)((char *)bufferp[0] - (char *)buffer);
  memcpy(next,buffer,length);
  next+=length;bpos++;

  if(bpos*channels==BUFMULT*2) {
    p->playdata(bigbuf,(int)(next-bigbuf));
    next=bigbuf;bpos=0;
  }

  for (int i = 0; i < channels; ++i)
    bufferp[i] = buffer + i;
}


#endif
